"""Handle PirateBay alternative proxies."""
import bs4

from mariner import exceptions, proxyplugin


class PirateBayProxy(proxyplugin.ProxyPlugin):
    """PirateBay proxy list."""
    proxy_page_url = 'https://proxybay.github.io'

    @staticmethod
    def _parse(page: str) -> str:
        """Parse the page and return a list of pages.

        Args:
            page: Proxy list page to parse.

        Returns:
            Urls of proxy websites.
        """
        soup = bs4.BeautifulSoup(page, 'lxml')
        data = soup.select('table')[0].select('tr')[1:]

        for site in data:
            if PirateBayProxy._is_online(site):
                return site.select('td.site')[0].a.get('href')

        raise exceptions.NoProxyAvailable('No proxy available for PirateBay')

    @staticmethod
    def _is_online(soup: bs4.Tag) -> bool:
        """Check if proxy is online or not.

        Args:
            soup: BeautifulSoup for Proxy data.

        Returns:
            True if site is listed online."""
        status = soup.select('td.status')[0].img.get('src')
        return True if 'up' in status else False



# -*- coding: future_fstrings -*-
"""Module for searching torrents on KickAssTorrents."""
import asyncio
from typing import Dict, Iterator

import aiohttp
import async_timeout
import bs4

from mariner import torrent, trackerplugin


class KickAssTorrents(trackerplugin.TrackerPlugin):
    """Represents KickAssTorrents search engine."""

    search_url = 'https://katcr.co/katsearch/page/1/{title}'
    aliases = ['kat']

    async def get_cookie(self, url: str) -> Dict:
        """Get KickAssTorrents session ID cookie.

        Args:
            url: Search page to get the cookie from.

        Returns:
            KickAssTorrents session cookie.
        """
        with async_timeout.timeout(10):
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    cookie = response.cookies.popitem()
                    return {cookie[0]: cookie[1].value}

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            search_url = self.search_url.format(title=title)
            cookie = await self.get_cookie(search_url)
            page = await self.get(search_url, cookies=cookie)
        except (OSError, asyncio.TimeoutError):
            self.log.error('Cannot reach server at %s', search_url)
        return self._parse(page)

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        contents = soup.find('table', class_='torrents_table')
        try:
            for line in contents.find_all('tr')[1:]:
                torrent_ = line.find(
                    'div', class_='torrents_table__torrent_name')
                name = str(torrent_.find(
                    'a', class_='torrents_table__torrent_title').string)
                magnet = torrent_.find(
                    'a', {'title': 'Torrent magnet link'})['href']
                tracker = self.__class__.__name__
                size = line.find('td', {'data-title': 'Size'}).get_text()
                date = line.find('td', {'data-title': 'Age'}).get_text()
                raw_seeds = line.find('td', {'data-title': 'Seed'}).string
                seeds = self._parse_number(raw_seeds)
                raw_leeches = line.find('td', {'data-title': 'Leech'}).string
                leeches = self._parse_number(raw_leeches)
                yield torrent.Torrent(name, tracker, magnet=magnet, size=size,
                                      date=date, seeds=seeds, leeches=leeches)
        except AttributeError:
            self.log.debug("No results found")
            yield from ()



# -*- coding: future_fstrings -*-
"""Module for searching torrents on LinuxTracker."""
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class LinuxTracker(trackerplugin.TrackerPlugin):
    """Represents LinuxTracker search engine."""

    log = logging.getLogger(__name__)

    search_url = 'http://linuxtracker.org/index.php?page=torrents&search={title}'

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        content = soup.find_all('table', {'class': 'lista', 'width': '100%'})
        for torrent_ in content[4]:
            try:
                name = str(torrent_.font.a.string)
                tracker = self.__class__.__name__

                links = torrent_.find_all('td', {'align': 'right'})[0]
                magnet = links.find_all('a')[0]['href']
                stub = links.find_all('a')[1]['href']
                url = f'http://linuxtracker.org/{stub}'

                details = torrent_.find_all('tr')
                date = details[0].get_text().strip().split()[2]
                size = ' '.join(details[1].get_text().split()[1:])
                raw_seeds = details[2].get_text().split()[1]
                seeds = self._parse_number(raw_seeds)
                raw_leeches = details[3].get_text().split()[1]
                leeches = self._parse_number(raw_leeches)

                yield torrent.Torrent(name, tracker, torrent=url, magnet=magnet,
                                      date=date, size=size, seeds=seeds, leeches=leeches)
            except AttributeError:
                pass



# -*- coding: future_fstrings -*-
"""Module for searching torrents on LimeTorrents."""
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin
from mariner.proxies import limetorrents


class LimeTorrents(trackerplugin.ProxyTrackerPlugin):
    """Represents LimeTorrents search engine."""

    search_url = '{proxy}/search/all/{title}/seeds/1'

    def __init__(self):
        super().__init__()
        self.proxies = limetorrents.LimeTorrentsProxy()

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        try:
            contents = soup.select('table.table2')[0].select('tr')[1:]
            for content in contents:
                data = content.select('td')
                name = str(data[0].select('a')[1].string)
                tracker = self.__class__.__name__

                url = data[0].div.a.get('href')

                size = str(data[2].string)
                date = str(data[1].string.split('-')[0].strip())
                seeds = self._parse_number(data[3].string)
                leeches = self._parse_number(data[4].string)

                yield torrent.Torrent(
                    name,
                    tracker,
                    torrent=url,
                    size=size,
                    date=date,
                    seeds=seeds,
                    leeches=leeches)
        except IndexError:
            yield from ()



"""Base class for adding torrent tracker support."""
import abc
import asyncio
import logging
from typing import Iterator

from mariner import exceptions, mixins, torrent

Url = str
Page = str


class TrackerMeta(abc.ABCMeta, type):
    """Metaclass to check that Tracker plugins override search_url."""
    def __new__(mcs, name, bases, namespace, **kwargs):
        if abc.ABC not in bases:
            if not namespace.get('search_url'):
                raise exceptions.PluginError('You must define search_url')
        return type.__new__(mcs, name, bases, namespace)


class TrackerPlugin(mixins.GetPageMixin, abc.ABC, metaclass=TrackerMeta):
    """Represent a search engine."""
    log = logging.getLogger(__name__)
    user_agent = {'user-agent': 'Mariner Torrent Downloader'}
    search_url = ''  # To be overwritten by subclasses
    aliases = []  # Aliases for the tracker name

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            search_url = self.search_url.format(title=title)
            page = await self.get(search_url)
        except (OSError, asyncio.TimeoutError):
            self.log.error('Cannot reach server at %s', search_url)
        return self._parse(page)

    @abc.abstractmethod
    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:
        """Parse result page.

        Args:
            raw: Raw HTML page.

        Returns:
            List of torrents names with URLs and magnet links.

        """
        raise NotImplementedError

    @staticmethod
    def _parse_number(number: str) -> int:
        """Parse a number string from HTML page and return an interger.

        Args:
            number: Number string to parse.

        Return:
            Parsed number.
        """
        squashed = number.replace(' ', '')
        return int(squashed.replace(',', ''))


class ProxyTrackerPlugin(TrackerPlugin, abc.ABC):
    """Base class for trackers, that support alternative proxies.

    Attributes:
        proxies: Override with instance of ProxyPlugin.
    """
    default_proxy = ''  # To be overwritten by subclasses

    def __init__(self):
        super().__init__()
        self.proxies = None

    async def get_proxy(self) -> str:
        """Return a responding proxy.

        Returns:
            URL for responding proxy.
        """
        return await self.proxies.get_proxy()

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            proxy = await self.get_proxy()
            search_url = self.search_url.format(proxy=proxy, title=title)
            page = await self.get(search_url)
        except (OSError, asyncio.TimeoutError):
            self.log.error('Cannot reach server at %s', search_url)
        return self._parse(page)

    @abc.abstractmethod
    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:
        """Parse result page.

        Args:
            raw: Raw HTML page.

        Returns:
            List of torrents names with URLs and magnet links.

        """
        raise NotImplementedError



"""Handle LimeTorrents alternative proxies."""
import bs4

from mariner import exceptions, proxyplugin


class LimeTorrentsProxy(proxyplugin.ProxyPlugin):
    """LimeTorrents proxy list."""
    proxy_page_url = 'https://limetorrents-proxy.ga/'

    @staticmethod
    def _parse(page: str) -> str:
        """Parse the page and return a list of pages.

        Args:
            page: Proxy list page to parse.

        Returns:
            Urls of proxy websites.
        """
        soup = bs4.BeautifulSoup(page, 'lxml')
        data = soup.select('table.proxy-list')[0].select('tr')[1:]

        for site in data:
            if LimeTorrentsProxy._is_online(site):
                return site.select('td.plc-1')[0].a.get('href')

        raise exceptions.NoProxyAvailable(
            'No proxy available for LimeTorrents')

    @staticmethod
    def _is_online(soup: bs4.Tag) -> bool:
        """Check if proxy is online or not.

        Args:
            soup: BeautifulSoup for Proxy data.

        Returns:
            True if site is listed online."""
        status = soup.select('span.label')[0].string
        return True if 'Online' in status else False



# -*- coding: future_fstrings -*-
"""Represent a downloadable torrent."""
import datetime

from mariner import mixins, utils


class Torrent(mixins.ComparableMixin):  # pylint: disable=too-many-instance-attributes
    """Class representing a Torrent."""

    def __init__(self,
                 name: str,
                 tracker: str,
                 *,
                 torrent: str = None,
                 magnet: str = None,
                 size: str = 'Unknown',
                 seeds: int = -1,
                 leeches: int = None,
                 date: str = None) -> None:
        self.name = name
        self.tracker = tracker
        self.torrent = torrent
        self.magnet = magnet
        self.size = size
        self.seeds = seeds
        self.leeches = leeches
        self.date = date

    @property
    def _cmpkey(self) -> int:
        """Key to use for torrent comparison."""
        return self.seeds

    @property
    def date(self) -> datetime.date:
        """Upload date in structured format."""
        return self._date

    @date.setter
    def date(self, value: str) -> None:
        if isinstance(value, (datetime.date, type(None))):
            self._date = value  # pylint: disable=attribute-defined-outside-init
        else:
            # Defer importing maya, as it is a slow import
            import maya
            try:
                self._date = maya.when(value).date  # pylint: disable=attribute-defined-outside-init
            except ValueError:
                self._date = maya.parse(value).date  # pylint: disable=W0201
            except TypeError:
                self._date = None  # pylint: disable=W0201

    @property
    def filename(self) -> str:
        """Filename to use for the torrent."""
        return self.name if '.torrent' in self.name else f'{self.name}.torrent'

    def colored(self) -> 'Torrent':
        """Return Torrent with colored fields."""
        return Torrent(name=utils.yellow(self.name[:80]),
                       tracker=self.tracker,
                       torrent=self.torrent,
                       magnet=self.magnet,
                       size=self.size,
                       seeds=utils.green(self.seeds),
                       leeches=utils.red(self.leeches) if self.leeches is not None else None,
                       date=self.date)

    def __repr__(self) -> str:
        return f"Torrent({self.name}, {self.tracker}, {self.torrent}, \
    {self.magnet}, {self.size}, {self.seeds}, {self.leeches}, {self.date})"

    def __str__(self) -> str:
        return self.__repr__()



# -*- coding: future_fstrings -*-
"""Module for searching torrents on Nyaa."""
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class Nyaa(trackerplugin.TrackerPlugin):
    """Represents Nyaa.si search engine."""

    search_url = 'https://nyaa.si/?f=0&c=0_0&q={title}'

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        try:
            contents = soup.select('tr.default')
            for content in contents:
                data = content.select('td')
                name = str(data[1].a.string)
                tracker = self.__class__.__name__

                links = data[2].select('a')
                url_stub = links[0].get('href')
                url = f'https://nyaa.si{url_stub}'
                magnet = links[1].get('href')

                size = str(data[3].string)
                date = str(data[4].string)
                seeds = int(data[5].string)
                leeches = int(data[6].string)

                yield torrent.Torrent(
                    name,
                    tracker,
                    magnet=magnet,
                    torrent=url,
                    size=size,
                    date=date,
                    seeds=seeds,
                    leeches=leeches)
        except IndexError:
            yield from ()



"""Utility mixins used in Mariner."""
import aiohttp
import async_timeout


class ComparableMixin:  # pylint: disable=too-few-public-methods
    """Implement comparisons, using _cmpkey attribute."""

    def _compare(self, other, method):
        """Wrap rich comparison methods.

        Args:
            other: Object to compare to.
            method: Comparison method to use.

        Returns:
            Comparison function.
        """
        try:
            return method(self._cmpkey, other._cmpkey)  # pylint: disable=protected-access
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or returned different type,
            # so cannot make a comparison with "other".
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)


class GetPageMixin:  # pylint: disable=too-few-public-methods
    """Represent a web scraper."""

    async def get(self, url: str, headers: str = None, cookies: str = None) -> str:
        """Get the requested page.

        Args:
            url: Url of the page to get.

        Returns:
            Raw HTML page.
        """
        with async_timeout.timeout(10):
            async with aiohttp.ClientSession(headers=headers, cookies=cookies) as session:
                async with session.get(url) as response:
                    return await response.text()



# -*- coding: future_fstrings -*-
"""CLI config command."""
import logging
import pprint

from cliff import command

from mariner import utils


class Config(command.Command):
    """Show or update configuration."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument('--show', '-s', action='store_true',
                            help='Show the configuration')
        parser.add_argument(
            'key', nargs='?', help='Option to change', default=None)
        parser.add_argument('value', nargs='?', help='New value', default=None)
        return parser

    @staticmethod
    def _update_dict(dict_, key, value):
        """Update option in dictionary to a new value.

        Args:
            dict_: Dictionary to update.
            key: Option to update.
            value: New value.
        """
        dict_[key] = value if key in dict_ else (
            Config._update_dict(c, key, value) for c in dict_.values())

    def take_action(self, parsed_args):
        """Copy chosen magnet link to clipboard.

        Args:
            parsed_args: List of parsed arguments.
        """
        key = parsed_args.key
        value = parsed_args.value
        show_ = parsed_args.show
        if not show_ and not (key and value):
            raise ValueError('Provide key and value to update or use --show')
        if show_:
            self.log.info('Mariner configuration:')
            pprint.pprint(
                self.app.config._config)  # pylint: disable=protected-access
        else:
            self._update_dict(self.app.config, key, value)
            self.log.info(f'Updated {utils.green(key)} to {utils.green(value)}')
